#define _USE_MATH_DEFINES
#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <algorithm>

#define SQR(X) ((X)*(X))
#define DIST(x1, y1, x2, y2) (SQR((x1) - (x2)) + SQR((y1) - (y2)))
#define EPS 1e-9

const double cAngle = 1.4;
const double minDist = 0.1;
World *world;

//bool comp(CMonster &X, CMonster &Y)
//{
//	if (DIST(world->player.PositionX, world->player.PositionY, X.PositionX, X.PositionY) + 1e-9 <
//		DIST(world->player.PositionX, world->player.PositionY, Y.PositionX, Y.PositionY)) return false;
//	return true;
//}

int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	world = new World(argv[1], atoi(argv[2]));
	float alpha = 0.0;
	float dA = 0.02 / cAngle;
	//Infinite loop

	int fl = 0;

	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

		double x = 0, y = 0; 
		if (alpha < 3.9 * M_PI )
		{
			if (alpha > M_PI + 1e-9) dA = 0.01 / cAngle;
			for (int k = 2; k < 12; k++)
			if (alpha > k * M_PI + 1e-9) dA /= 1.2;
			if (alpha > 2.5 * M_PI && alpha < 3 * M_PI) dA *= 1.4;
			if (alpha > 3 * M_PI && alpha < 3.9 * M_PI) dA /= 1.4;
			alpha += dA;
			x = cos(alpha);
			y = sin(alpha);
		}
		else
		{
			if (fl == 0) {fl = 1; alpha = 6*M_PI; dA = 0.006;}
			alpha += dA;
			x = cos(alpha);
			y = sin(alpha);
			/*x = y = 0;
			for (int i = 0; i < world->MAXMONSTERS; i++)
			{
				if (world->monsters[i].ClassName < 0) continue;
				x += (world->monsters[i].PositionX - world->player.PositionX);
				y += (world->monsters[i].PositionY - world->player.PositionY);
			}*/
		}
		world->run(x, y);

		int num = -1;
		for (int k = 0; k < world->MAXMONSTERS; k++)
		{
			if (world->monsters[k].ClassName < 0) continue;
			if (num == -1) num = k;
			if (DIST(world->player.PositionX, world->player.PositionY, world->monsters[k].PositionX, world->monsters[k].PositionY) + 1e-9 <
				DIST(world->player.PositionX, world->player.PositionY, world->monsters[num].PositionX, world->monsters[num].PositionY)) num = k;
		}
		world->shoot(world->monsters[num].PositionX - world->player.PositionX,  world->monsters[num].PositionY - world->player.PositionY);
		if (DIST(world->player.PositionX, world->player.PositionY, world->monsters[num].PositionX, world->monsters[num].PositionY) > minDist + EPS)
		{
			num = -1;
			for (int k = 0; k < world->MAXMONSTERS; k++)
			{
				if (world->monsters[k].ClassName < 0) continue;
				if (num == -1) num = k;
				if (world->monsters[k].HitPoints < world->monsters[num].HitPoints) num = k;
			}
			world->shoot(world->monsters[num].PositionX - world->player.PositionX,  world->monsters[num].PositionY - world->player.PositionY);
		}
		if (alpha < 3.5 * M_PI || alpha > 4 * M_PI) continue;
		double sx = 0, sy = 0;
		int count = 0;
		for (int i = 0; i < world->MAXMONSTERS; i++)
		{
			if (world->monsters[i].ClassName < 0) continue;
			count++;
			sx += world->monsters[i].PositionX;
			sy += world->monsters[i].PositionY;
		}
		sx /= count;
		sy /= count;
		world->shoot(sx - world->player.PositionX,  sy - world->player.PositionY);
	}

	return 0;
}

