#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>

const double r = 0.4;

const double pi = acos(-1.0);
const double eps = 1e-9;
const double INF = 1e100;

#define sqr(a) ((a)*(a))
#define clr(a) memset(a, 0, sizeof(a))

bool dead[1024];
int revive[1024];
int MaxTime;     
int last = -1;


bool isSniperMode(int cadr)
{
//	return 1;
	return (cadr <= 000) || (cadr >= 1750 && cadr <= 3200) || (cadr >= 4250 && cadr <= 5000);
}

double _acos(double x)
{
	if (fabs(x) > 1 - eps)
		if (x > 0)
			return 0;
		else
			return pi;
	return acos(x);
}

struct Point {
	double x, y;

	Point (double _x, double _y)
	{
		x = _x;
		y = _y;
	}

	Point (CPlayer & pl)
	{
		x = pl.PositionX;
		y = pl.PositionY;
	}

	Point (CMonster & m)
	{
		x = m.PositionX;
		y = m.PositionY;
	}

	double abs() const
	{
		return sqrt(x * x + y * y);
	}

	Point operator - (const Point & p) const
	{
		return Point(x - p.x, y - p.y);
	}

	double distTo(const Point & p) const
	{
		return (*this - p).abs();
	}

	double operator % (const Point & p) const
	{
		return x * p.x + y * p.y;
	}

	double getAngle(const Point & p) const
	{
		return _acos(*this % p / abs() / p.abs());
	}
};

int maxShooter = 6000;

int countSniper(World * world)
{
	int cnt = 0;
	for (int i = 0; i < world-> MAXMONSTERS; ++i)
		if (world->monsters[i].ClassName >= 0)
		{
			if (world -> monsters[i].BulletDamage > 0)
				++cnt;
		}
	return cnt;
}

double getDist(CPlayer & _pl, CMonster & _m)
{
	Point pl (_pl);
	Point m (_m);
	Point v (_pl.RunX, _pl.RunY);
	double d = pl.distTo(m);

	double ang = v.getAngle(m - pl);

	if (_m.BulletDamage > 0)
	{
		fprintf(stderr, "%.10lf\n", _m.BulletDamage);
		fflush(stderr);
	}

	double k = (last <= 7500 || last >= 7800) ? 20 : 1;
	if (last <= 2000)
		return d + ang / 10;
	double ans = d + ang / k;
	if (ans < 0.15)
		ans -= 100;

	if ((last >= 5500 && last <= maxShooter) || (1500 <= last && last <= 1800))
	{
		return d + _m.HitPoints * 2;
	}

	fprintf(stderr, "hit: %.10lf\n", _m.HitPoints);
	ans += _m.HitPoints / 3 /*/ _m.MaxHitPoints*/;
	if (last <= 5000)
		ans -= _m.BulletDamage * 40;
	else
		ans -= _m.BulletDamage * 20;

	return ans;
}

double shoot(World * world, CMonster m)
{
	double x = world -> player.PositionX;
	double y = world -> player.PositionY;

	double x0 = m.PositionX;
	double y0 = m.PositionY;

	double vx = m.VelocityX;
	double vy = m.VelocityY;
	double t;

	for (int i = 0; i < 10; ++i)
	{

		t = Point(x, y).distTo(Point(x0, y0)) / world -> player.BulletVelocity;
		
	 	x0 = m.PositionX + vx * t;
		y0 = m.PositionY + vy * t;
	}

	world -> shoot(x0 - x, y0 - y);

	return t;
}

int shootTarget = -1;
int shootRemain = 0;

void move(World * world)
{
	int cadr = world -> player.Time;
	if (cadr == last)
		return;
    last = cadr;

    if (last >= 5500 && last <= 5510)
    {
    	if (countSniper(world) <= 5)
    		maxShooter = 6500;
    	else
    		maxShooter = 6000;
    }

	double x = world -> player.PositionX;
	double y = world -> player.PositionY;

	if (sqr(x - 0.5) + sqr(y - 0.5) < 0.2 * 0.2)
		world -> run(0, 1);
	else
	{
		double alpha = atan2(y - 0.5, x - 0.5);

		if (last <= 7900)
			alpha += 0.02;
		else
			alpha -= 0.02;

		double newX = cos(alpha) * r + 0.5;
		double newY = sin(alpha) * r + 0.5;

		world -> run(newX - x, newY - y);
	}

	bool fl = 0;
	int p = -1; 
	double bestD = INF;
	for (int i = 0; i < world-> MAXMONSTERS; ++i)
	{
		if (revive[i] < cadr)
			dead[i] = false;

		if (world->monsters[i].ClassName >= 0)
		{
			if (dead[i])
				continue;

				//Shoot the bastard! (Shoot in the direction of the monster)
			double curD = getDist(world -> player, world -> monsters[i]);

			if (!fl || bestD > curD)
			{
				bestD = curD;
				p = i;
				fl = 1;
			}
		}
		else
			dead[i] = false;
	}


	if (isSniperMode(cadr) && shootRemain == 0 && fl)
	{
		shootTarget = p;
		shootRemain = world -> monsters[p].HitPoints / world -> player.BulletDamage + 1;
		if (cadr >= 5250)
			shootRemain *= 2;
		dead[p] = true;
		revive[p] = int(1e6);
	}

	fl |= (shootRemain != 0);
		

	if (fl)
	{
//		printf("%.10lf\n", world -> player.GunCooldownRemaining);
		if (isSniperMode(cadr))
		{
			double t = shoot(world, world -> monsters[shootTarget]);
			if (world -> player.GunCooldownRemaining < 1 / 60.)
			{
				--shootRemain;
				if (shootRemain == 0)
					revive[shootTarget] = cadr + t * 60 + 2;
			}
		}
		else
			shoot(world, world -> monsters[p]);
	}
	else
	{
		world -> shoot(0, 0);
		shootRemain = 0;
	}
}

int main(int argc, char** argv) {
    freopen("log.txt", "w", stderr);
	
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	clr(dead);

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));
	float alpha = 0.0;

	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf remain: %d\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY, shootRemain);
		fprintf(stderr, "frame : %d, HP: %.2lf x: %.2lf y:%.2lf remain: %d\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY, shootRemain);

		move(world);

	}

	return 0;
}
