#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <vector>
#include <algorithm>

using namespace std;

typedef double rtype;

const int MAXANGLE = 256;
const int MA4 = MAXANGLE / 4;
rtype danger[MAXANGLE];
const rtype MINWALLDIST = 0.1;
const rtype MINLLAWDIST = 1.0 - MINWALLDIST;
const rtype MINMOBDIST = 0.1;
const rtype PI = 3.1415926;
const rtype PI2 = PI * 2.0;
int debugged = 0;

const int LASTMOBS = 3;
CMonster *lastmob[LASTMOBS] = {0, 0, 0};
int bulletcount[LASTMOBS] = {0, 0, 0};
int lastmobcnt = 0;


float get_distance(float x1, float y1, float x2, float y2) {
	float dx, dy;
	dx = x1 - x2;
	dy = y1 - y2;
	return sqrt(dx * dx + dy * dy);
}

inline rtype ang2rad(int ang){
	return (PI2 * ang) / MAXANGLE;
}

inline int rad2ang(rtype rad){
	return static_cast<int>(floor(MAXANGLE * rad / PI2)) % MAXANGLE;
}

bool isangleok(rtype x, rtype y, int ang){
	return true;
	if (x<MINWALLDIST){
		if (y<MINWALLDIST){
			return (ang<=MA4);
		} else if(y>MINLLAWDIST){
			return (ang>=3*MA4);
		} else{
			return (x >= 3*MA4) || (x <= MA4);
		}
	} else if (x>MINLLAWDIST){
		if (y<MINWALLDIST){
			return (ang>=MA4) && (ang <= 2*MA4);
		} else if(y>MINLLAWDIST){
			return (ang>=2*MA4) && (ang <= 3*MA4);
		} else{
			return (x >= MA4) || (x <= 3*MA4);
		}
	} else{
		if (y<MINWALLDIST){
			return (ang <= 2*MA4);
		} else if(y>MINLLAWDIST){
			return (ang>=2*MA4);
		} else{
			return true;
		}
	}
}

inline void addMob(int type, rtype x, rtype y, rtype q, rtype x0, rtype y0, rtype* vx, rtype*vy) {
	rtype dx, dy, dist, dist2;
	dx = x - x0;
	dy = y - y0;
	dist2 = dx*dx + dy*dy;
	if (dist2 < MINMOBDIST) {
		dist2 = MINMOBDIST;
	}
	dist = sqrt(dist2);
	switch(type) {
		case 2: q = q / (dist2); break;
		case 1:	q = q / dist; break;
		case 4: q = q / dist2 * dist2; break;
	}
	dx /= dist;
	dy /= dist;
	*vx -= dx * q;
	*vy -= dy * q;
}

void get_move(World* w, rtype* ansx, rtype* ansy) {
	rtype vx = 0, vy = 0, dx = 0, dy = 0;
	CMonster * const mobend = w->monsters + w->MAXMONSTERS;
	for (CMonster *mob = w->monsters; mob < mobend; ++ mob){
		if (mob->ClassName <= 0)
			continue;
		addMob(2, mob->PositionX, mob->PositionY, (mob->BiteDamage + mob->BulletDamage), w->player.PositionX, w->player.PositionY, &vx, &vy);
	}
	rtype tmp = vx;
	vx -= vy / 4.0;
	vy += tmp / 4.0;
	dx = (1 - w->player.PositionX);
	dy = (1 - w->player.PositionY);
	vx -= 2 / dx;
	vx += 2 / w->player.PositionX;
	vy -= 2 / dy;
	vy += 2 / w->player.PositionY;
	*ansx = vx;
	*ansy = vy;
}

void get_my_move(World* w, rtype* ansx, rtype* ansy) {
	memset(danger, 0, MAXANGLE*sizeof(rtype));
	CMonster * const mobend = w->monsters + w->MAXMONSTERS;
	for (CMonster *mob = w->monsters; mob < mobend; ++ mob){
		if (mob->ClassName <= 0)
			continue;
		rtype dx = mob->PositionX - w->player.PositionX;
		rtype dy = mob->PositionY - w->player.PositionY;
		rtype dist = sqrt(dx*dx + dy*dy);
		if (dist < MINMOBDIST)
			dist = MINMOBDIST;
		rtype dangle = atan2(dy, dx);
		int angle = rad2ang(dangle);
		danger[angle] += (mob->BiteDamage + mob->BulletDamage) / dist; //TODO best function using attack value
	}
	rtype mindanger = 100500.0;
	int minindex = 64;
	for (int i=MAXANGLE-1; i>=0; --i){
	//for (int i=0; i<MAXANGLE; ++i){
		if (isangleok(w->player.PositionX, w->player.PositionY, i) && (danger[i] < mindanger)){
			minindex = i;
			mindanger = danger[i];
		}
	}
	if (debugged == 0){
		debugged = 1;
		for (int i=0; i<MAXANGLE; ++i)
			printf("%f ", danger[i]);
	}
	*ansx = cos(ang2rad(minindex));
	*ansy = sin(ang2rad(minindex));
}


void get_shoot(World* world, rtype* ansx, rtype* ansy) {
	rtype cdist, mdist = 2;
	rtype minhp = -1;
	CMonster *mon, *smon;
	for (int j = 0; j < world->MAXMONSTERS; j++) {
		mon = &world->monsters[j];
		if (mon->ClassName > 0) {
			cdist = get_distance(mon->PositionX, mon->PositionY, world->player.PositionX, world->player.PositionY);
			if (cdist < mdist) {
				mdist = cdist;
			}
		}
	}
	if (mdist == 2) {
		*ansx = 0;
		*ansy = 0;
		return;
	}
	for (int j = 0; j < world->MAXMONSTERS; j++) {
		mon = world->monsters+j;
		for (int k=0; k<LASTMOBS; ++k){
			if ((lastmob[k] == mon) && (bulletcount[k] <= 0))
				continue;
		}
		if (mon->ClassName > 0) {
			cdist = get_distance(mon->PositionX, mon->PositionY, world->player.PositionX, world->player.PositionY);
			if (cdist <= mdist + mon->Radius * 2 && (mon->HitPoints < minhp || minhp < 0)) {
				smon = mon;
			}
		}
	}
	bool isinlast = false;
	for (int k=0; k<LASTMOBS; ++k){
		if (lastmob[k] == smon){
			isinlast = true;
			if(world->player.GunCooldownRemaining <= 0.05)
				--(bulletcount[k]);
			break;
		}
	}
	if (!isinlast){
		++lastmobcnt;
		lastmobcnt %= LASTMOBS;
		lastmob[lastmobcnt] = smon;
		bulletcount[lastmobcnt] = static_cast<int>(floor(smon->HitPoints / world->player.BulletDamage)) + 2;
	}
	rtype m, g, n, d, t1, t2, ux1, ux2, ux, uy;
	n = world->player.PositionX - smon->PositionX;
	m = world->player.PositionY - smon->PositionY;
	g = n * smon->VelocityY -  m * smon->VelocityX;
	d = sqrt(m * m + n * n - g * g);
	t1 = 1 / (m * m + n * n);
	ux1 = -(m * g + n * d) * t1;
	ux2 = -(m * g - n * d) * t1;
	t1 = n / (smon->VelocityX - ux1);
	t2 = n / (smon->VelocityX - ux2);
	if (t1 > 0) {
		uy = smon->VelocityY - m / t1;
		ux = ux1;
	} else {
		uy = smon->VelocityY - m / t2;
		ux = ux2;
	}
	*ansx = ux;
	*ansy = uy;
}

int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));
	rtype xx, yy;
	
	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		/*printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);*/

		get_move(world, &xx, &yy);
		world->run(xx, yy);
		get_shoot(world, &xx, &yy);
		world->shoot(xx, yy);
	}

	return 0;
}
