#define _USE_MATH_DEFINES

#include "world.h"
#include <cmath>
#include <ctime>
#include <algorithm>
#include <iostream>

using namespace std;

int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3)
	{
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));
	int vects[8][2] = { { 0, 1 }, { 1, 1 }, { 1, 0 }, { 1, -1 }, { 0, -1 }, { -1, -1 }, { -1, 0 }, { -1, 1 } };
	int phase = 0;

	//Infinite loop
	while (true)
	{
		//Send our turn and receive the update of the game world
		world->makeMove();

		const float maxr = 0.16;
		int neightbors[128] = { 0 };
		CMonster *closest = NULL;
		float closd;

		//Choose the first monster that is still alive
		for (int j = 0; j < world->MAXMONSTERS; j++)
		{
			if (world->monsters[j].ClassName < 0)
			{
				continue;
			}

			if (vects[phase][0] * world->monsters[j].PositionX
				+ vects[phase][1] * world->monsters[j].PositionY < 0.7)
			{
				continue;
			}

			float distX = world->player.PositionX - world->monsters[j].PositionX;
			float distY = world->player.PositionY - world->monsters[j].PositionY;
			float dist = distX * distX + distY * distY;

			if ((closest == NULL || dist < closd) && dist < 0.16)
			{
				closest = &world->monsters[j];
				closd = dist;
			}
		}

		if (closest == NULL)
		{
			for (int j = 0; j < world->MAXMONSTERS; j++)
			{
				if (world->monsters[j].ClassName < 0)
				{
					continue;
				}

				float distX = world->player.PositionX - world->monsters[j].PositionX;
				float distY = world->player.PositionY - world->monsters[j].PositionY;
				float dist = distX * distX + distY * distY;

				if (world->monsters[j].VelocityX == 0 && world->monsters[j].VelocityY == 0)
				{
			        if (sqrt(dist) / world->player.BulletVelocity < world->monsters[j].GunCooldownRemaining)
			        {
						if (closest == NULL || dist < closd)
						{
							closest = &world->monsters[j];
							closd = dist;
						}
	               	}
				}
			}

		}

		if (closest == NULL)
		{
			for (int j = 0; j < world->MAXMONSTERS; j++)
			{
				if (world->monsters[j].ClassName < 0)
				{
					continue;
				}

				float distX = world->player.PositionX - world->monsters[j].PositionX;
				float distY = world->player.PositionY - world->monsters[j].PositionY;
				float dist = distX * distX + distY * distY;

				if (closest == NULL || dist < closd)
				{
					closest = &world->monsters[j];
					closd = dist;
				}
			}
		}

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

		const float delta = 0.03;
		const float rad = 0.3;
		bool done = false;

		switch (phase)
		{
		case 0:
			if (world->player.PositionY > 1 - rad)
			{
				done = true;
			}
			break;

		case 1:
			if (world->player.PositionY > 1 - delta)
			{
				done = true;
			}
			break;

		case 2:
			if (world->player.PositionX > 1 - rad)
			{
				done = true;
			}
			break;

		case 3:
			if (world->player.PositionX > 1 - delta)
			{
				done = true;
			}
			break;

		case 4:
			if (world->player.PositionY < rad)
			{
				done = true;
			}
			break;

		case 5:
			if (world->player.PositionY < delta)
			{
				done = true;
			}
			break;

		case 6:
			if (world->player.PositionX < rad)
			{
				done = true;
			}
			break;

		case 7:
			if (world->player.PositionX < delta)
			{
				done = true;
			}
			break;
		}

		if (done)
		{
			phase = (phase + 1) % 8;
			cout << "Changed phase to " << phase << "!" << endl;
		}

		world->run(vects[phase][0], vects[phase][1]);

   		if (closest != NULL)
		{
//			float t = (world->player.PositionX - closest->PositionX) / (closest->VelocityX - closest->VelocityX);

//			float targetX = world->player.PositionX;

			world->shoot(closest->PositionX - world->player.PositionX,
					     closest->PositionY - world->player.PositionY);
   		}
   	}

	return 0;
}
