import java.io.IOException;
import java.util.Arrays;
import java.util.Random;

import static java.lang.Math.*;
import static java.util.Arrays.fill;
import static java.util.Arrays.sort;

class Task {
	static World world;

	public static void main(String[] args) throws IOException {
		world = new World(args[0], Integer.parseInt(args[1]));
		new Task().run();
	}
	

	int frame;
	float time;

	Random rnd = new Random(1238721);
	Point dir = new Point();
	Point shootVec = new Point();
	Point[] oldMonsterPos;

	Strategy naive = new NaiveStrategy();

	MoveSquarePattern moveSquarePattern = new MoveSquarePattern();
	MoveCirclePattern moveCirclePattern = new MoveCirclePattern();
	
	int[] cntShoot = new int[World.MAXMONSTERS];
	
	int oldMCnt = -1;
	
	void run() {
		world.makeMove();
		frame = world.player.Time;
		oldMonsterPos = new Point[World.MAXMONSTERS];
		for (int i = 0; i < oldMonsterPos.length; i++)
			oldMonsterPos[i] = new Point(world.monsters[i].PositionX,
					world.monsters[i].PositionY);
		while (true) {
			world.makeMove();

			int dt = world.player.Time - frame;
			if (dt == 0)
				continue;

			time = dt / 60.0f;

			naive.run();
			world.run(dir.x, dir.y);
			world.shoot(shootVec.x, shootVec.y);

			frame = world.player.Time;
			for (int i = 0; i < oldMonsterPos.length; i++)
				oldMonsterPos[i].set(world.monsters[i].PositionX,
						world.monsters[i].PositionY);
		}
	}

	class Strategy {
		void run() {
		}
	}

	class NaiveStrategy extends Strategy {
		int CHANGE_AIM_RATE = 15;
		int CHANGE_DIR_RATE = 1;
		int CLEAR_SHOOT_CNT_RATE = 120;
		
		float MIN_BORDER_DIST = 0.2f;

		float STUPID_SHOOT_BARIER = 0.0f;

		int IT = 32;
		int dirTime = CHANGE_DIR_RATE - 1;
		
		int shootTime = 0;


		Point escapeDir = new Point();

		float px, py;

		float mcx, mcy;

		int oldAim;
		
		int shootsPerAim = -1;

		@Override
		void run() {
			px = world.player.PositionX;
			py = world.player.PositionY;

			makeShot();
		}

		// void changeDir() {
		// dir.set(0.5f - rnd.nextFloat(), 0.5f - rnd.nextFloat());
		// }

		void makeShot() {
			calcMonsterCenter();
			calcEscapeDir(mcx, mcy);
			dir.set(escapeDir.x, escapeDir.y);
			
			
			int id = getAim();
//			System.out.println(id);
			
			if (id == -1)
				return;
			
			if (world.player.GunCooldownRemaining <= 1f / 60 + 1e-9)
				cntShoot[id]++;
			

			calcShootVector(id);
		}

		private void calcMonsterCenter() {
			mcx = 0;
			mcy = 0;
			int mcnt = 0;
			for (int i = 0; i < world.monsters.length; i++) {
				CMonster monster = world.monsters[i];
				if (monster.ClassName > 0) {
					mcx += monster.PositionX;
					mcy += monster.PositionY;
					mcnt++;
				}
			}

			if (mcnt == 0)
				return;

			mcx /= mcnt;
			mcy /= mcnt;
		}

		private void calcEscapeDir(float mx, float my) {

			if (++dirTime == CHANGE_DIR_RATE) {
				
				int curCntM = getMonsters(0, 0, 1);
				if (oldMCnt != -1 && curCntM > oldMCnt) {
					escapeDir.set(0.5f - px, 0.5f - py);
					CHANGE_DIR_RATE = 30;
					
				} else {
				
				CHANGE_DIR_RATE = 1;

//				moveSquarePattern.move(escapeDir);
				
				moveCirclePattern.move(escapeDir);

				float borderDist = getPMinDistToBorder();
				if (borderDist < MIN_BORDER_DIST) {
					escapeDir.rotate(PI / 2);
				}
				}
				
			dirTime = 0;
			
			oldMCnt = curCntM;

			}
			
			dir.set(escapeDir.x, escapeDir.y);

		}

		float calcShootVector(int mid) {

			float mx = world.monsters[mid].PositionX;
			float my = world.monsters[mid].PositionY;

			float vx = (mx - oldMonsterPos[mid].x) / time;
			float vy = (my - oldMonsterPos[mid].y) / time;

			if (len(vx, vy) < STUPID_SHOOT_BARIER) {
				shootVec.set(mx - px, my - py);
				return -1;
			}

			float lt = 0.0f;
			float rt = 1e4f;
			for (int it = 0; it < IT; it++) {
				float md = 0.5f * (lt + rt);
				float cx = mx + vx * md;
				float cy = my + vy * md;
				if (dist(px, py, cx, cy) > md * world.player.BulletVelocity)
					lt = md;
				else
					rt = md;
			}
			float ret = 0.5f * (lt + rt);
			// System.out.println(ret);

			float cx = mx + vx * ret;
			float cy = my + vy * ret;
			shootVec.set(cx - px, cy - py);
			return ret;
		}

		int getAim() {
			shootTime++;
			
			if (shootTime == CLEAR_SHOOT_CNT_RATE) {
				shootTime = 0;
				fill(cntShoot, 0);
			}
			
			for (int i = 0; i < World.MAXMONSTERS; i++) {
				CMonster monster = world.monsters[i];
				if (monster.ClassName < 0)
					cntShoot[i] = 0;
			}
			
			int ret = -1;
			for (int i = 0; i < World.MAXMONSTERS; i++) {
				CMonster monster = world.monsters[i];
				if (monster.ClassName > 0) {
					if (cntShoot[i] < ceil(monster.MaxHitPoints / world.player.BulletDamage) * 1.5f && (ret == -1
							|| better(monster, world.monsters[ret])))
						ret = i;
				}
			}
			oldAim = ret;

			return ret;
		}

		boolean better(CMonster m1, CMonster m2) {
			return distPM(world.player, m1) < distPM(world.player, m2);
		}

		float getPMinDistToBorder() {
			// float dist = 2;
			// for (int i = 0; i <= 1; i++)
			// for (int j = 0; j <= 1; j++)
			// dist = min(dist, dist(px, py, i, j));
			float dist = px;
			dist = min(dist, py);
			dist = min(dist, 1 - px);
			dist = min(dist, 1 - py);
			return dist;
		}

	}

	private int getMonsters(float x, float y, float side) {
		int ret = 0;
		for (CMonster monster : world.monsters)
			if (monster.ClassName > 0 && contains(x, y, side, monster))
				ret++;
		return ret;
	}
	private boolean contains(float x, float y, float side, CMonster monster) {
		return monster.PositionX >= x && monster.PositionX <= x + side
				&& monster.PositionY >= y && monster.PositionY <= y + side;
	}

	float len(float x, float y) {
		return (float) sqrt(x * x + y * y);
	}

	float dist(float x1, float y1, float x2, float y2) {
		return (float) sqrt(sqr(x1 - x2) + sqr(y1 - y2));
	}

	double distPM(CPlayer p, CMonster m) {
		return sqrt(sqr(p.PositionX - m.PositionX)
				+ sqr(p.PositionY - m.PositionY));
	}

	float sqr(float x) {
		return x * x;
	}

	class MovePattern {

		float px, py;

		void move(Point dir) {
			px = world.player.PositionX;
			py = world.player.PositionY;
		}

	}

	class MoveCirclePattern extends MovePattern {

		void move(Point dir) {
			super.move(dir);

			float ex = px - 0.5f;
			float ey = py - 0.5f;

			if (ex == 0 && ey == 0) {
				ex = -1;
				ey = 0;
			}

			dir.set(ex, ey);
		}
		
		
	}

	class MoveSquarePattern extends MovePattern {

		int MAX_MONSTER_CHANGE = 2;

		int cnt = 8;

		float oldRx, oldRy;
		int oldCntM = -1;

		@Override
		void move(Point dir) {
			super.move(dir);

			float side = 1f / cnt;

			if (oldCntM != -1
					&& getMonsters(oldRx, oldRy, side) - oldCntM <= MAX_MONSTER_CHANGE)
				return;

			float best = -1;
			float rx = -1, ry = -1;
			int rCntM = -1;
			for (float x = 0; x < 1f; x += side)
				for (float y = 0; y < 1f; y += side) {
					int cntM = getMonsters(x, y, side);

					float cur = dist(px, py, x + side / 2, y + side / 2)
							/ (cntM + 1);
					if (cur > best) {
						best = cur;
						rCntM = cntM;
						rx = x;
						ry = y;
					}
				}

			oldRx = rx;
			oldRy = ry;
			oldCntM = rCntM;

			System.out.println(rx + " " + ry + " " + best);
			dir.set(rx + side / 2 - px, ry + side / 2 - py);
		}

	}

	class Point {
		float x;
		float y;

		Point() {
		}

		public void rotate(double ang) {
			double cos = cos(ang);
			double sin = sin(ang);
			float nx = (float) (x * cos - y * sin);
			float ny = (float) (x * sin + y * cos);
			set(nx, ny);
		}

		Point(float x, float y) {
			this.set(x, y);
		}

		void set(float x, float y) {
			this.x = x;
			this.y = y;
		}
	}
}
