#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <iostream>
#include <vector>
using namespace std;

struct Queue
{
	vector <World> arr;
	int now;
	int size;
	int Max;
	Queue(int Max = 10):Max(Max)
	{
		now = 0;
		size=0;
	}
	void Add(World &w)
	{
		if (arr.size() < Max)
		{
			arr.push_back(w);
			size++;
			return;
		}
		arr[now] = w;
		now = (now+1)%Max;
	}
	World& operator [](int x)
	{
		return arr[(now-1-x+2*Max)%Max];
	}


};


struct V2
{
	double x,y;
	V2(double x =0, double y = 0):x(x),y(y){}
	
	double Size()
	{
		return sqrt(x*x+y*y);
	}
	bool Zero()
	{
		return Size() < 1e-9;
	}

	V2 Rotate(double al)
	{
		V2 w;
		w.x = cos(al)*x+sin(al)*y;
		w.y = -sin(al)*x+cos(al)*y;
		return w;
	}

	V2 Norm()
	{
		return V2(x/Size(),y/Size());
	}
};

V2 operator +(const V2& a, const V2& b)
{
	return V2(a.x+b.x, a.y+b.y);
}
V2 operator -(const V2& a, const V2& b)
{
	return V2(a.x-b.x, a.y-b.y);
}

double operator *(const V2& a, const V2& b)
{
	return a.x*b.x + a.y*b.y;
}

double operator ^(const V2& a, const V2& b)
{
	return a.x*b.y- a.y*b.x;
}

V2 operator /(const V2& a, double d)
{
	return V2(a.x/d, a.y/d);
}

V2 operator *( double d, const V2& a)
{
	return V2(a.x*d, a.y*d);
}

Queue q(10);
World *world;
double BORDER = 0.2;
int dx[]={-1,0,1,0};
int dy[]={0,1,0,-1};
double view_angle1 = 0.5;
double view_angle2 = 0.5;
double VIEW_PRIORITY1 = 10;
double VIEW_PRIORITY2 = 5;
double DODGE = 0.5;
int last_monster = -2;

const int PHASE1 = 1850;
const int PHASE2 = 3200;
const double PI = 3.141592653589793238462;
const int PHASE2_SHOOT_MOD = 1;
const int PHASE1_SHOOT_MOD = 2;
vector <int> last_shoot;	

V2 GetPlayer()
{
	return  V2(world->player.PositionX,world->player.PositionY);
}

V2 GetMonster(int idx)
{
	return  V2(world->monsters[idx].PositionX,world->monsters[idx].PositionY);
}


void MakeMove()
{
		world->makeMove();

		//printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
		//		world->player.Time, world->player.HitPoints,
		//		world->player.PositionX, world->player.PositionY);
		//q.Add(*world);
}

V2 GetWeightCenter()
{
	V2 v;
	int cnt = 0;
	for (int i = 0; i < world->MAXMONSTERS; i ++)
		if (world->monsters[i].ClassName >= 0) 
		{
			v = v+V2(world->monsters[i].PositionX,world->monsters[i].PositionY);
			cnt ++;
		}
	if (cnt == 0)
		return v;
	v = v/cnt;
	return v;
}


V2 BorderCorrection(V2 v, V2 c)
{
	V2 p = GetPlayer();
	
	if (p.x < BORDER || 1-p.x < BORDER)
	{
		V2 w;
		w.y = v.x;
		w.x = -v.y;
		if (!(p.y < BORDER || 1-p.y < BORDER))
			w.x =0; //  /= 10;
		v = w;

	}else
	if (p.y < BORDER || 1-p.y < BORDER)
	{
		V2 w;
		w.y = v.x;
		w.x = -v.y;
		w.y =0;//  /= 10;
		v = w;
	}


	return v;
}



int FindMonster(V2 v)
{
	int ans = -1;
	bool good = false;
	double res;
	V2 p = GetPlayer();
	
	V2 v1 = v.Rotate(view_angle1);
	V2 v2 = v.Rotate(-view_angle1);


	for (int i = 0; i < world->MAXMONSTERS; i ++)
		if (world->monsters[i].ClassName >= 0) 
		{
			V2 m(world->monsters[i].PositionX,world->monsters[i].PositionY);
			bool tgood = false;

			double r =  (m-p).Size();

			if (((m-p)^v1)*((m-p)^v2) < 0 && ((m-p)*v) > 0)
				r /= VIEW_PRIORITY1;
			//r*= world->monsters[i].MaxHitPoints;
			//else
			//	if (((m-p)^w1)*((m-p)^w2) < 0 && ((m-p)*v) > 0)
			//		r /= VIEW_PRIORITY2;
	
			//if (world->monsters[i].HitPoints/world->monsters[i].MaxHitPoints < 0.5)
			//	r /= 2;
			if (ans == -1)
			{
				ans = i;
				res = r;
				continue;
			}

			if (r < res || ans == last_monster)
			{
				ans = i;
				res = r;
			}
		}
	last_monster = ans;
	return ans;
}


int FindStableMonster()
{
	int ans = -1;
	int res = 0;
	V2 p = GetPlayer();
	
	for (int i = 0; i < world->MAXMONSTERS; i ++)
		if (world->monsters[i].ClassName >= 0) 
		{
			V2 m(world->monsters[i].PositionX,world->monsters[i].PositionY);
			V2 v(world->monsters[i].VelocityX,world->monsters[i].VelocityY);
			if (v.Zero())
			{
				if (ans == -1 || (res > (m-p).Size() && i != last_monster))
				{
					ans = i;
					res = (m-p).Size();
				}
			}
		}
	last_monster = ans;
	return ans;
}

V2 ShootDirection(int idx)
{
	if (idx == -1)
		return V2(0,0);
	V2 p = GetPlayer();
	
	V2 m(world->monsters[idx].PositionX,world->monsters[idx].PositionY);
	return m-p;
}

V2 TryDodge()
{
	V2 p = GetPlayer();
	
	int worst=-1;
	V2 v(0,0);
	for (int i = 0; i < world->MAXMISSILES; i ++)
		if (world->missiles[i].ClassName >= 0) 
		{
			V2 m(world->missiles[i].PositionX,world->missiles[i].PositionY);
			V2 w(world->missiles[i].VelocityX,world->missiles[i].VelocityY);
			if (fabs((p-m).Norm()*w.Norm()-1) < 1e-5)
			{
				worst = i;
				v = w.Rotate(PI/2).Norm();
				break;
			}
		}
	return DODGE*v;
}

double GetT(int i, V2 m)
{
	//V2 m = GetMonster(i);
	V2 p = GetPlayer();
	double v = world->player.BulletVelocity;
	V2 vb(world->monsters[i].VelocityX, world->monsters[i].VelocityY);
	double xb = m.x, yb = m.y, x0 = p.x, y0 = p.y;
	double a = (vb.x*vb.x+vb.y*vb.y-v*v);
	double b = 2*(xb-x0)*vb.x+2*(yb-y0)*vb.y;
	double c = (xb-x0)*(xb-x0) + (yb-y0)*(yb-y0);
	if(fabs(a) < 1e-9)
		return -c/b;
	return max((-b+sqrt(b*b-4*a*c))/(2*a), (-b-sqrt(b*b-4*a*c))/(2*a));
}


V2 GetShootDirect2()
{
	for (int i = 0; i < world->MAXMONSTERS; i ++)
		if (world->monsters[i].ClassName >= 0) 
		{
			int need = world->monsters[i].HitPoints / world->player.BulletDamage + PHASE2_SHOOT_MOD;
			if (GetMonster(i).x < 0 || GetMonster(i).y < 0 || GetMonster(i).x > 1 || GetMonster(i).y > 1)
				cout << "FAIL";
			double glt = 0;
			bool ok = true;
			/*for(int j = 0; j < need; j++)
			{
				double t = j * world->player.GunCooldownPeriod;
				V2 now(GetMonster(i).x + world->monsters[i].VelocityX * t, GetMonster(i).y + world->monsters[i].VelocityY * t);
				/*if(now.x > 1 || now.y > 1 || now.x < 0 || now.y < 0)
				{
					ok = false;
					break;
				}
				double tans = GetT(i, now);
				if(tans < 0)
				{
					ok = false;
					break;
				}
				tans += t;
				now = V2(GetMonster(i).x + world->monsters[i].VelocityX * tans, GetMonster(i).y + world->monsters[i].VelocityY * tans);
				/*if(now.x > 1 || now.y > 1 || now.x < 0 || now.y < 0)
				{
					ok = false;
					break;
				}
			}*/
			//if(ok)
			{
				double t = GetT(i, GetMonster(i));
				if (world->player.Time - last_shoot[i] < 10)
					continue;
				if(t < 0)
					continue;
				last_shoot[i] =  world->player.Time +3;
				V2 m = GetMonster(i);
				V2 np = V2(m.x + world->monsters[i].VelocityX * t, m.y + world->monsters[i].VelocityY * t);
				//cout << "OK" << endl;
				return V2(np-GetPlayer());
			}
		}
	return V2();
}

void Phase()
{


	int shoots=0;
	int phase = 1;
	int cur_monster=-1;
	last_shoot.resize(world->MAXMONSTERS,0);
	while (true)
	{
		if (world->player.Time > PHASE1)
			phase = 2;
		if (world->player.Time > PHASE2)
			phase = 3;

		MakeMove();
		V2 v,c,m;
		V2 p(world->player.PositionX,world->player.PositionY);
		c = GetWeightCenter();
		v = p-c;
		v = BorderCorrection(v,c);
		world->run(v.x,v.y);

		//if (phase == 2)
		//	v = TryDodge();
		//if (phase == 3)
		//	v = v + TryDodge();
		if (phase == 1 || phase == 3)
		{
			if (cur_monster == -1 || shoots == 0 || world->monsters[cur_monster].HitPoints <= 0)
			{
				//int t_monster = -1;
				//if (phase == 2)
				//	t_monster = FindStableMonster();
				//if (t_monster!= -1)
				//	cur_monster = t_monster;
				//else
					cur_monster = FindMonster(v);

				if (cur_monster != -1)
					shoots = world->monsters[cur_monster].HitPoints/(world->player.BulletDamage)+PHASE1_SHOOT_MOD;
			}
			
			m = ShootDirection(cur_monster);
			if (world->player.GunCooldownRemaining == 0)
				shoots--;
			world->shoot(m.x,m.y);
		}
		else
		{
				//cout << "OK";
				V2 w = GetShootDirect2();
				world->shoot(w.x,w.y);
		}
	}
}

int main(int argc, char** argv) {
	//Check for enough parameters
	freopen("error.txt", "w", stderr);
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}
	world = new World(argv[1], atoi(argv[2]));
	
	Phase();
	return 0;
}
