#define _CRT_SECURE_NO_DEPRECATE

#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <vector>
#include <algorithm>

using namespace std;

#define pb push_back
#define mp make_pair
#define fi(a,b) for (i=a;i<=b;i++)
#define fj(a,b) for (j=a;j<=b;j++)
#define fo(a,b) for (o=a;o<=b;o++)
#define fdi(a,b) for (i=a;i>=b;i--)
#define fdj(a,b) for (j=a;j>=b;j--)
#define fdo(a,b) for (o=a;o>=b;o--)
#define SIZE(x) (int)x.size()
#define LEN(x) (int)x.length()
#define ZERO(x) memset(x,0,sizeof(x))

#define INF 1000000000
#define UPDATE 30

double mhp[10000];
double bhp[10000];
int mup[10000];
int bup[10000];

double movex, movey;

double shootx, shooty;

double pi = acos(-1.);

World *world;

vector <double> angle;

double eps = 1e-7;

double sqr(double x) {
	return x * x;
}

double Dist(double x1, double y1, double x2, double y2) {
	return sqrt(sqr(x2 - x1) + sqr(y2 - y1));
}

double Len(double x, double y) {
	return Dist(0, 0, x, y);
}

inline
double PlayX() {
	return world->player.PositionX;
}

inline
double PlayY() {
	return world->player.PositionY;
}

double MonsterAngle(int x) {
	return atan2(world->monsters[x].PositionY - world->player.PositionY, 
			world->monsters[x].PositionX - world->player.PositionX);
}

double MonsterDist(int x) {
	return Dist(world->monsters[x].PositionX, world->monsters[x].PositionY,
			PlayX(), PlayY());
}

double BulletDist(int x) {
	return Dist(world->missiles[x].PositionX, world->missiles[x].PositionY,
			PlayX(), PlayY());
}

bool AnlgeBetween(double a1, double a2, double a) {
	if (a1 <= a && a <= a2) return true;
	if (a1 <= a + 2 * pi && a + 2 * pi <= a2) return true;
	if (a1 <= a - 2 * pi && a - 2 * pi <= a2) return true;
	return false;
}

double ToX(double x) {
	return x - world->player.PositionX;
}

double ToY(double y) {
	return y - world->player.PositionY;
}

double ToX(double x, double y) {
	return x - world->player.PositionX;
}

double ToY(double x, double y) {
	return y - world->player.PositionY;
}

double W;

bool ShootToBullets() {
	double mind;
	int minv;
	int i;
	double t;
	double bx, by;
	double d;
	double bt;
	double bv;
	double vel;
	mind = INF;
	minv = -1;
	fi(0, world->MAXMISSILES - 1) {
		if (world->missiles[i].ClassName == -1 || world->missiles[i].ClassName == 200) continue;
		if (bhp[i] < eps) continue;
		if (BulletDist(i) < mind) {
			mind = BulletDist(i);
			minv = i;
		}
	}
	if (minv == -1) return false;

	vel = world->player.BulletVelocity;

	bt = BulletDist(minv) / vel;

	bv = 0;

	for (t = 0; t <= 50; t += 0.02) {
		d = Dist(PlayX(), PlayY(), world->missiles[minv].PositionX + world->missiles[minv].VelocityX * t,
			world->missiles[minv].PositionY + world->missiles[minv].VelocityY * t);
		if (fabs(d / vel - t) < bt) {
			bt = fabs(d / vel - t);
			bv = t;
		}
	}

	bx = world->missiles[minv].PositionX + world->missiles[minv].VelocityX * bv;
	by = world->missiles[minv].PositionY + world->missiles[minv].VelocityY * bv;

	bhp[minv] -= world->player.BulletDamage;

	shootx = ToX(bx);
	shooty = ToY(by);

	printf("Shoot to bullet!\n");

	return true;
}

double Angle(double x1, double y1, double x2, double y2) {
	double res;
	res = fabs(atan2(y1, x1) - atan2(y2, x2));
	res = min(res, fabs(atan2(y1, x1) + 2 * pi - atan2(y2, x2)));
	res = min(res, fabs(atan2(y1, x1) - 2 * pi - atan2(y2, x2)));
	return res;
}

void ShootToClosest() {
	if (world->player.GunCooldownRemaining > eps) return;
	double mind;
	int minv;
	int i;
	double t;
	double bx, by;
	double d;
	double bt;
	double bv;
	double vel;
	double minhp;
	if (ShootToBullets()) return;

	mind = INF;
	minv = -1;
	fi(0, world->MAXMONSTERS - 1) {
		if (world->monsters[i].ClassName == -1) continue;
		if (mhp[i] < eps) continue;

		if (Angle(movex, movey, world->monsters[i].PositionX - PlayX(), world->monsters[i].PositionY - PlayY()) > pi / 10) continue;

		if (MonsterDist(i) < mind) {
			mind = MonsterDist(i);
			minv = i;
		}
	}

	if (minv == -1) {
		minhp = INF;
		fi(0, world->MAXMONSTERS - 1) {
			if (world->monsters[i].ClassName == -1) continue;
			if (mhp[i] < eps) continue;
			minhp = min(minhp, mhp[i]);
		}
		fi(0, world->MAXMONSTERS - 1) {
			if (world->monsters[i].ClassName == -1) continue;
			if (mhp[i] < eps) continue;
			if (mhp[i] > minhp + eps) continue;

			if (MonsterDist(i) < mind) {
				mind = MonsterDist(i);
				minv = i;
			}
		}

	}

	if (minv == -1) return;

	vel = world->player.BulletVelocity;

	bt = MonsterDist(minv) / vel;

	bv = 0;

	for (t = 0; t <= 50; t += 0.02) {
		d = Dist(PlayX(), PlayY(), world->monsters[minv].PositionX + world->monsters[minv].VelocityX * t,
			world->monsters[minv].PositionY + world->monsters[minv].VelocityY * t);
		if (fabs(d / vel - t) < bt) {
			bt = fabs(d / vel - t);
			bv = t;
		}
	}

	bx = world->monsters[minv].PositionX + world->monsters[minv].VelocityX * bv;
	by = world->monsters[minv].PositionY + world->monsters[minv].VelocityY * bv;

	mhp[minv] -= world->player.BulletDamage;

	shootx = ToX(bx);
	shooty = ToY(by);

}

int Count(double x1, double y1, double x2, double y2) {
	int i;
	int res;
	res = 0;
	fi(0, world->MAXMONSTERS - 1) {
		if (world->monsters[i].ClassName == -1) continue;
		if (x1 <= world->monsters[i].PositionX && world->monsters[i].PositionX <= x2 &&
			y1 <= world->monsters[i].PositionY && world->monsters[i].PositionY <= y2) res++;		
	}
	return res;
}

void ImproveDirection() {
	double vx, vy;
	double l, r;
	int i;
	l = Len(movex, movey);
	r = 28;
	movex /= l;
	movey /= l;	
	movex *= r;
	movey *= r;
	fi(0, world->MAXMONSTERS - 1) {
		if (world->monsters[i].ClassName == -1) continue;
		vx = -(world->monsters[i].PositionX - PlayX());
		vy = -(world->monsters[i].PositionY - PlayY());
		
		l = Len(vx, vy);
		r = Len(vx, vy);

		if (l > 0.25) continue;

		vx /= l;
		vy /= l;

		vx *= pow((1 / r), 2);
		vy *= pow((1 / r), 2);
		
		movex += vx;
		movey += vy;
	}
}

int selectedquarter;

int q[10][10];
int n = 4;

int DX[] = {-1,0,1,-1,1,-1,0,1};
int DY[] = {-1,-1,-1,0,0,1,1,1};

bool Field(int x, int y) {
	if (x <= 1 || x >= n) return false;
	if (y <= 1 || y >= n) return false;
	return true;
}

int T;

bool disimpr;

void Inside() {
	int qx, qy;
	int sx, sy;
	int bx, by;
	int i, j;

	if (selectedquarter != -1) {
		qx = (selectedquarter - 1) % n + 1;
		qy = (selectedquarter - 1) / n + 1;

		movex = ToX(qx * W - W / 2, qy * W - W / 2);
		movey = ToY(qx * W - W / 2, qy * W - W / 2);

		if (fabs(PlayX() - (qx * W - W / 2)) < W / 4 && fabs(PlayY() - (qy * W - W / 2)) < W / 4) {
			printf("Free (center)\n");
			selectedquarter = -1;
		} else {

			T--;
			if (T == 0) {
				printf("Free (time)\n");
				selectedquarter = -1;
			}

		}

		if (!disimpr) ImproveDirection();
		ShootToClosest();
		return;
	}

	sx = (int)(PlayX() / W) + 1;
	sy = (int)(PlayY() / W) + 1;

	int x2, y2;
	int vv;
	double vd;

	vv = 1000000;	
	bx = 1;
	by = 1;

	fi(0, 7) {
		x2 = sx + DX[i];
		y2 = sy + DY[i];
		if (!Field(x2, y2)) continue;
		if (q[y2][x2] < vv) {
			vv = q[y2][x2];
			bx = x2;
			by = y2;
			disimpr = 0;
			T = 85;
		}
	}

	if (q[by][bx] >= 4) {

		printf("GO!\n");

		fj(2, n - 1) {
			fi(2, n - 1) {
				if (i == sx && j == sy) continue;
				if (q[j][i] < vv) {
					bx = i;
					by = j;
					vv = q[j][i];
					disimpr = 1;
					T = 260;
				}
			}
		}

	}

	double d;

	if (q[by][bx] == 0) {
		printf("Empty square\n");
		vd = Dist(sx, sy, bx, by);
		fj(2, n - 1) {
			fi(2, n - 1) {
				if (q[j][i] != 0) continue;
				d = Dist(sx, sy, bx, by);
				if (d < vd) {
					vd = d;
					bx = i;
					by = j;
				}
			}
		}
	}

	if (q[sy][sx] == 0 && q[by][bx] == 0 && q[4][4] == 0 && q[5][4] == 0 && q[4][5] == 0 && q[3][4] == 0 && q[4][3] == 0) {
		if (sx != 4 || sy != 4) {
			printf("Center!\n");
			bx = 4;
			by = 4;
		}
	}


	printf("Go from %d %d to %d %d\n", sx, sy, bx, by);

	selectedquarter = (by - 1) * n + bx;

	


	/*alpha = 0;
	fi(0, SIZE(angle) - 2) {
		if (angle[i + 1] - angle[i] > alpha) {

			if (cos((angle[i] + angle[i + 1]) / 2) < 0) {
				dx = -1;
			} else {
				dx = 1;
			}

			if (sin((angle[i] + angle[i + 1]) / 2) < 0) {
				dy = -1;
			} else {
				dy = 1;
			}

			if (dx == -1 && dy == -1 && Count(0, 0, W, W) > 0) continue;
			if (dx == 1 && dy == 1 && Count(1 - W, 1 - W, 1, 1) > 0) continue;
			if (dx == -1 && dy == 1 && Count(0, 1 - W, W, 1) > 0) continue;
			if (dx == 1 && dy == -1 && Count(1 - W, 0, 1, W) > 0) continue;

			alpha = angle[i + 1] - angle[i];
			alpha1 = angle[i];
			alpha2 = angle[i + 1];
		}
	}
	if (alpha == 0) return;

	if (cos((alpha1 + alpha2) / 2) < 0) {
		dx = -1;
	} else {
		dx = 1;
	}

	if (sin((alpha1 + alpha2) / 2) < 0) {
		dy = -1;
	} else {
		dy = 1;
	}

	if (dx == -1 && dy == -1) selectedquarter = 2;
	if (dx == 1 && dy == -1) selectedquarter = 1;
	if (dx == -1 && dy == 1) selectedquarter = 3;
	if (dx == 1 && dy == 1) selectedquarter = 4;

	ShootToClosest();
	fi(0, world->MAXMONSTERS - 1) {
		if (world->monsters[i].ClassName == -1) continue;
		if (alpha1 <= MonsterAngle(i) && MonsterAngle(i) <= alpha2) {
			movex = cos((alpha1 + alpha2) / 2);
			movey = sin((alpha1 + alpha2) / 2);
			return;
		}
	}*/
}

void Move() {
//	double alpha;
	int i, j;
	//ImproveDirection();
	//ShootToClosest();

	ZERO(q);

	fj(1, n) {
		fi(1, n) {
			q[j][i] = Count((i - 1) * W, (j - 1) * W, i * W, j * W);
		}
	}

	Inside();

	ShootToClosest();

	/*
	angle.clear();
	fi(0, world->MAXMONSTERS - 1) {
		if (world->monsters[i].ClassName == -1) continue;
		angle.pb(MonsterAngle(i));
	}
	if (SIZE(angle) == 0) return;
	sort(angle.begin(), angle.end());
	angle.pb(angle[0]);
	alpha = 0;
	fi(0, SIZE(angle) - 2) {
		alpha = max(alpha, angle[i + 1] - angle[i]);
	}
	if (alpha >= pi) {
		Outside();
	} else {
		Inside();
	}*/
}

double G = 0.15;

int mcolor[10000];

void Update() {
	int i;
	fi(0, world->MAXMONSTERS - 1) {
		if (world->monsters[i].ClassName == -1) {
			mcolor[i] = 0;
			if (mhp[i] > 0) {
				mhp[i] = 0;
			}
		} else {
			if (mhp[i] < eps && mcolor[i] == 0) {
				mcolor[i] = 1;
				mup[i] = UPDATE;
				mhp[i] = world->monsters[i].HitPoints + 2 * world->player.BulletDamage;
			} else if (mcolor[i] == 1) {
				if (mup[i]) {
					mup[i]--;
				} else {
					mhp[i] = world->monsters[i].HitPoints;
					mup[i] = UPDATE;
				}
			}
		}
	}
	fi(0, world->MAXMISSILES - 1) {
		if (world->missiles[i].ClassName == -1) {
			mcolor[i] = 0;
			if (bhp[i] > 0) {
				bhp[i] = 0;
			}
		} else {
			if (bhp[i] < eps && mcolor[i] == 0) {
				mcolor[i] = 1;
				bup[i] = UPDATE;
				bhp[i] = world->missiles[i].HitPoints + world->player.BulletDamage;
			} else if (mcolor[i] == 1) {
				if (bup[i]) {
					bup[i]--;
				} else {
					bhp[i] = world->missiles[i].HitPoints;
					bup[i] = UPDATE;
				}
			}
		}
	}
}

int main(int argc, char** argv) {
	int i;
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
			return 10;
	}

	//Create the world data (and establish connection with the server)
	world = new World(argv[1], atoi(argv[2]));

	selectedquarter = -1;

	n = 8;
	W = 1. / n;

	fi(0, world->MAXMONSTERS - 1) {
		mup[i] = UPDATE;
	}

	fi(0, world->MAXMISSILES - 1) {
		bup[i] = UPDATE;
	}

	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		/*printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);*/

		//Run in circles (set direction)

		movex = 0;
		movey = 0;
		shootx = 0;
		shooty = 0;

		fflush(stdout);

		Update();

		Move();

		if (PlayX() <= G && movex < 0) movex = 0;
		if (PlayX() >= 1 - G && movex > 0) movex = 0;
		if (PlayY() <= G && movey < 0) movey = 0;
		if (PlayY() >= 1 - G && movey > 0) movey = 0;

		world->run((float)movex, (float)movey);
		world->shoot((float)shootx, (float)shooty);

		//Choose the first monster that is still alive
		/*for ( int j = 0; j < world->MAXMONSTERS; j++)
			if (world->monsters[j].ClassName >= 0) {
				//Shoot the bastard! (Shoot in the direction of the monster)
				world->shoot(world->monsters[j].PositionX - world->player.PositionX,
						   world->monsters[j].PositionY - world->player.PositionY);
				break;
			}*/
	}

	return 0;
}
