#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>

const int DEL_MONSTER = 7;
const int DEL_SHOOT = 6;
const double GR = 0.01;

const int INF = 1000*1000*1000;

const int PCNT = 8;
const int POS[PCNT][2] = 
{
	{1,0},
	{-1,0},
	{1,1},
	{0,1},
	{-1,1},
	{1,-1},
	{0,-1},
	{-1,1}
};

int ** init(int sz)
{
	int ** r = new int*[sz];
	for(int i=0; i<sz; ++i)
		r[i] = new int[sz];
	return r;
}

int ** mat_monster = init(DEL_MONSTER);
int ** mat_shoot = init(DEL_SHOOT);
World * world;

void clear(int ** mat, const int sz)
{
	for(int i=0; i<sz; ++i)
		for(int j=0; j<sz; ++j)
			mat[i][j] = 0;
}

const bool in(const int i, const int j, const  int k, const double d)
{
	double x1 = i*d;
	double y1 = j*d;
	double x2 = i*d+d;
	double y2 = j*d+d;

	if ( world->monsters[k].ClassName>=0 )
	{
		double x  = world->monsters[k].PositionX;
		double y  = world->monsters[k].PositionY;
		return x>=x1 && x<=x2 && y>=y1 && y<=y2;
	}
	else
	{
		return false;
	}
}

void calc(int ** mat, const  int sz)
{
	double d = 1.0/sz;
	for(int i=0; i<sz; ++i)
		for(int j=0; j<sz; ++j)
			for(int k=0; k<world->MAXMONSTERS; ++k)
				if (in(i, j, k, d))
					mat[i][j]++;
}

int cnt(int x, int y, int ** const mat, const int sz)
{
	if (x<0 || x>=sz || y<0 || y>=sz )
		return INF;
	else
		return mat[x][y];
}

int main(int argc, char** argv) {

	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}
	world = new World(argv[1], atoi(argv[2]));

	while (true) {
		world->makeMove();

		{
			clear(mat_monster, DEL_MONSTER);
			calc(mat_monster, DEL_MONSTER);
		
			double d_monster = 1.0/DEL_MONSTER;
			int x = (int)floor(world->player.PositionX/d_monster);
			int y = (int)floor(world->player.PositionY/d_monster);
			int bp = -1;
			int bc = INF;
			for(int i=0; i<8; ++i)
			{
				int t = cnt(x+POS[i][0], y+POS[i][1], mat_monster, DEL_MONSTER);
				if (t<bc)
				{
					bc = t;
					bp = i;
				}
			}
			world->run(POS[bp][0], POS[bp][1]);
		}

		{
			/*
			clear(mat_shoot, DEL_SHOOT);
			calc(mat_shoot, DEL_SHOOT);

			int bpi = 0;
			int bpj = 0;
			int bc = 0;

			for(int i=0; i<DEL_SHOOT; ++i)
				for(int j=0; j<DEL_SHOOT; ++j)
					if (mat_monster[i][j]>bc) {
						bc = mat_monster[i][j];
						bpi = i;
						bpj = j;
					}
*/
			double xx = -1;
			double yy = -1;
			double r = INF;
			double d= 1.0/DEL_SHOOT;
			for(int k=0; k<world->MAXMONSTERS; ++k) {
				double x = world->monsters[k].PositionX;
				double y = world->monsters[k].PositionY; 
				double rr = sqrt(pow(world->player.PositionX-x,2) + pow(world->player.PositionY-y, 2));
				//if (in(bpi, bpj, k, d) && r>rr)
				if ( r>rr)
				{
					r = rr;
					xx = x;
					yy = y;
				}
			}

			xx -= world->player.PositionX;
			yy -= world->player.PositionY;

			//printf("%lf %lf\n", xx, yy);

			world->shoot(xx, yy);
		}
	}

	return 0;
}

