#include "Director.h"
#include "GameWorld.h"

#define GUY(STR) STR##_CLASSNAME, STR##_RADIUS, STR##_HITPOINTS, STR##_VELOCITY, STR##_ACCELERATION
#define WRANGED(STR) STR##_PERIOD, STR##_RADIUS, STR##_VELOCITY, STR##_DAMAGE, STR##_HITPOINTS
#define WMELEE(STR) STR##_PERIOD, STR##_DAMAGE
#define DRONEAI(STR) mon->radius * STR##_AWARENESS_RATIO, STR##_THINK_PERIOD, STR##_PLAYER_IMPORTANCE, STR##_RANDOM_IMPORTANCE
#define DUCKAI(STR) mon->radius * STR##_AWARENESS_RATIO
#define WALLAI(STR) DUCKAI(STR)
#define DODGERAI(STR) mon->radius * STR##_AWARENESS_RATIO, STR##_THINK_PERIOD, STR##_PLAYER_IMPORTANCE, STR##_RANDOM_IMPORTANCE, STR##_MISSILE_IMPORTANCE

//----------------------------------PLAYER------------------------------
//player parameters
static const int PLAYER_CLASSNAME = 0;
static const double PLAYER_RADIUS = 0.01;
static const double PLAYER_HITPOINTS = 1.0;	//DO NOT FORGET TO CHANGE TO 1
static const double PLAYER_VELOCITY = 0.15;
static const double PLAYER_ACCELERATION = 1.0;
//player gun parameters
static const double PLAYER_GUN_PERIOD = 0.1;
static const double PLAYER_GUN_RADIUS = 0.001;
static const double PLAYER_GUN_VELOCITY = 0.5;
static const double PLAYER_GUN_DAMAGE = 0.056;
static const double PLAYER_GUN_HITPOINTS = 0.11;

//---------------------------------DRONE-------------------------------
//drone monster parameters
static const int DRONE_MONSTER_CLASSNAME = 1;
static const double DRONE_MONSTER_RADIUS = 0.008;
static const double DRONE_MONSTER_HITPOINTS = 0.1;
static const double DRONE_MONSTER_VELOCITY = 0.1;
static const double DRONE_MONSTER_ACCELERATION = 0.3;
//drone monster ai parameters
static const double DRONE_AI_AWARENESS_RATIO = 5.0;
static const double DRONE_AI_THINK_PERIOD = 0.2;
static const double DRONE_AI_PLAYER_IMPORTANCE = 1.0;
static const double DRONE_AI_RANDOM_IMPORTANCE = 0.3;
//drone monster teeth parameters
static const double DRONE_TEETH_PERIOD = 1.0;
static const double DRONE_TEETH_DAMAGE = 0.03;

//---------------------------------SHOOTER-------------------------------
//shooter monster parameters
static const int SHOOTER_MONSTER_CLASSNAME = 2;
static const double SHOOTER_MONSTER_RADIUS = 0.005;
static const double SHOOTER_MONSTER_HITPOINTS = 0.03;
static const double SHOOTER_MONSTER_VELOCITY = 0.05;
static const double SHOOTER_MONSTER_ACCELERATION = 0.5;
//shooter monster ai parameters
static const double SHOOTER_AI_AWARENESS_RATIO = 5.0;
static const double SHOOTER_AI_THINK_PERIOD = 1.0;
static const double SHOOTER_AI_PLAYER_IMPORTANCE = 0.5;
static const double SHOOTER_AI_RANDOM_IMPORTANCE = 0.3;
//shooter monster teeth parameters
static const double SHOOTER_TEETH_PERIOD = 1.0;
static const double SHOOTER_TEETH_DAMAGE = 0.003;
//shooter gun parameters
static const double SHOOTER_GUN_PERIOD = 0.5;
static const double SHOOTER_GUN_RADIUS = 0.001;
static const double SHOOTER_GUN_VELOCITY = 0.2;
static const double SHOOTER_GUN_DAMAGE = 0.03;
static const double SHOOTER_GUN_HITPOINTS = 0.03;

//---------------------------------WALL-------------------------------
//wall monster parameters
static const int WALL_MONSTER_CLASSNAME = 3;
static const double WALL_MONSTER_RADIUS = 0.01;
static const double WALL_MONSTER_HITPOINTS = 0.5;
static const double WALL_MONSTER_VELOCITY = 0.06;
static const double WALL_MONSTER_ACCELERATION = 1.0;
//wall monster ai parameters
static const double WALL_AI_AWARENESS_RATIO = 0.0;
//wall monster teeth parameters
static const double WALL_TEETH_PERIOD = 0.1;
static const double WALL_TEETH_DAMAGE = 0.05;

//---------------------------------SPEEDY-------------------------------
//speedy monster parameters
static const int SPEEDY_MONSTER_CLASSNAME = 4;
static const double SPEEDY_MONSTER_RADIUS = 0.008;
static const double SPEEDY_MONSTER_HITPOINTS = 0.05;
static const double SPEEDY_MONSTER_VELOCITY = 0.25;
static const double SPEEDY_MONSTER_ACCELERATION = 0.0;
//speedy monster ai parameters
static const double SPEEDY_AI_AWARENESS_RATIO = 0.0;
//shooter gun parameters
static const double SPEEDY_GUN_PERIOD = 2.0;
static const double SPEEDY_GUN_RADIUS = 0.001;
static const double SPEEDY_GUN_VELOCITY = 3.0;
static const double SPEEDY_GUN_DAMAGE = 0.02;
static const double SPEEDY_GUN_HITPOINTS = 0.02;

//---------------------------------DODGER-------------------------------
//dodger monster parameters
static const int DODGER_MONSTER_CLASSNAME = 5;
static const double DODGER_MONSTER_RADIUS = 0.006;
static const double DODGER_MONSTER_HITPOINTS = 0.25;
static const double DODGER_MONSTER_VELOCITY = 0.2;
static const double DODGER_MONSTER_ACCELERATION = 0.5;
//dodger monster ai parameters
static const double DODGER_AI_AWARENESS_RATIO = 10.0;
static const double DODGER_AI_THINK_PERIOD = 0.0; //always
static const double DODGER_AI_PLAYER_IMPORTANCE = -0.1; //stay away
static const double DODGER_AI_RANDOM_IMPORTANCE = 0.3;
static const double DODGER_AI_MISSILE_IMPORTANCE = 10.0; //dodge
//dodger gun parameters
static const double DODGER_GUN_PERIOD = 0.2;
static const double DODGER_GUN_RADIUS = 0.0001;
static const double DODGER_GUN_VELOCITY = 0.3;
static const double DODGER_GUN_DAMAGE = 0.02;
static const double DODGER_GUN_HITPOINTS = 0.02;


double2 RandomFieldPosition() {
	return double2(trandom(), trandom()) * FIELD_SIZE;
}

double2 RandomBorderPosition(int mask = -1) {
	double2 rnd = double2(trandom(), trandom());
	rnd.y *= 0.1;
	int tt;
	do {
		tt = rand() & 3;
	} while (!(mask & (1<<tt)));
	     if (tt == 0) rnd = double2(rnd.x, 1.0 + rnd.y);
	else if (tt == 2) rnd = double2(rnd.x, -rnd.y);
	else if (tt == 1) rnd = double2(1.0 + rnd.y, rnd.x);
	else if (tt == 3) rnd = double2(-rnd.y, rnd.x);
	return rnd * FIELD_SIZE;
}

void Director::Init() {
	Player *pl = new Player(double2(1.0, 1.0) * (FIELD_SIZE/2), GUY(PLAYER));
	pl->SetRangedWeapon(new RangedWeapon(WRANGED(PLAYER_GUN)));
	owner->BindEntity(0, pl);
}

void Director::SpawnDrone(double2 position) {
	Monster *mon = new Monster(position, GUY(DRONE_MONSTER));
	mon->SetMeleeWeapon(new MeleeWeapon(WMELEE(DRONE_TEETH)));
	mon->SetAI(new DroneAI(mon, DRONEAI(DRONE_AI)));
	owner->BindEntity(owner->AllocateMonster(), mon);
}

void Director::SpawnShooter(double2 position) {
	Monster *mon = new Monster(position, GUY(SHOOTER_MONSTER));
	mon->SetMeleeWeapon(new MeleeWeapon(WMELEE(SHOOTER_TEETH)));
	mon->SetRangedWeapon(new RangedWeapon(WRANGED(SHOOTER_GUN)));
	mon->SetAI(new DroneAI(mon, DRONEAI(SHOOTER_AI)));
	owner->BindEntity(owner->AllocateMonster(), mon);
}

void Director::SpawnSpeedy(double2 direction) {
	double2 position = RandomFieldPosition() - ForceVectorNorm(direction, 1.0);
	ForceToField(position, 0.0);
	Monster *mon = new Monster(position, GUY(SPEEDY_MONSTER));
	mon->velocity = direction * SPEEDY_MONSTER_VELOCITY;
	mon->SetRangedWeapon(new RangedWeapon(WRANGED(SPEEDY_GUN)));
	mon->SetAI(new WallAI(mon, WALLAI(SPEEDY_AI), direction));
	owner->BindEntity(owner->AllocateMonster(), mon);
}

void Director::SpawnWall(int num) {
	static const double2 pts[] = {double2(0, 0)
		, double2(1, 0)
		, double2(1, 1)
		, double2(0, 1)
		, double2(0, 0)};

	double dist = WALL_MONSTER_RADIUS * 4.0;
	double2 pfirst = pts[num] * FIELD_SIZE;
	double2 pdiff = (pts[num+1] - pts[num]);
	double2 vel(-pdiff.y, pdiff.x);
	pdiff = ForceVectorNorm(pdiff, 2.0 * WALL_MONSTER_RADIUS + 2.0 * PLAYER_RADIUS);
	int cnt = int(FIELD_SIZE / !pdiff) + 1;

	ControlPopulation(MAX_MONSTER_COUNT - cnt);

	for (int i = 0; i<cnt; i++) {
		double2 pos = pfirst + pdiff * i;
		Monster *mon = new Monster(pos, GUY(WALL_MONSTER));
		mon->SetMeleeWeapon(new MeleeWeapon(WMELEE(WALL_TEETH)));
		mon->SetAI(new WallAI(mon, WALLAI(WALL_AI), vel));
		mon->SetMoveDirection(vel);
		owner->BindEntity(owner->AllocateMonster(), mon);
	}
}

void Director::SpawnDodger(double2 position) {
	Monster *mon = new Monster(position, GUY(DODGER_MONSTER));
	mon->SetRangedWeapon(new RangedWeapon(WRANGED(DODGER_GUN)));
	mon->SetAI(new DodgerAI(mon, DODGERAI(DODGER_AI)));
	owner->BindEntity(owner->AllocateMonster(), mon);
}


//do not affect walls
void Director::ControlPopulation(int maxcnt) {
	static std::vector<Monster *> alive;
	alive.clear();
	alive.reserve(MAX_MONSTER_COUNT);

	int rem = 0;
	for (int i = 0; i<MAX_MONSTER_COUNT; i++) if (owner->monsters[i]) {
		if (owner->monsters[i]->classname == WALL_MONSTER_CLASSNAME) rem++;
		else alive.push_back(owner->monsters[i]);
	}
	int kill = rem + alive.size() - maxcnt;
	if (kill < 0) return;
	if (kill > alive.size()) kill = alive.size();

	static std::vector<Monster *> dead;
	dead.clear();
	dead.reserve(MAX_MONSTER_COUNT);

	for (int i = 0; i<kill; i++) {
		int j = rand() % alive.size();
		dead.push_back(alive[j]);
		std::swap(alive[j], alive.back());
		alive.pop_back();
	}

	for (int i = 0; i<dead.size(); i++) owner->DestroyEntity(dead[i]);
}



void Director::PassTime(double deltatime) {
	gametime += deltatime;

	if (gametime < 20.0) {
		static Cooldown cool(8.0, 0.0);
		static int k = 50;
		cool.PassTime(deltatime);
		if (cool.TimeToAct()) {
			ControlPopulation(MAX_MONSTER_COUNT - k);
			for (int i = 0; i<k; i++) SpawnDrone(RandomBorderPosition());
			k *= 2;
		}
	}

	if (gametime > 30.0 && gametime < 55.0) {
		static Cooldown cool(0.5, 0.0);
		static bool init = false;
		if (!init) { init = true;  ControlPopulation(0); }
		cool.PassTime(deltatime);                      
		if (cool.TimeToAct()) SpawnSpeedy(RandomDirection());
	}

	if (gametime > 53.0 && gametime < 60.0) {
		static Cooldown cool(0.1, 0.0);
		cool.PassTime(deltatime);
		if (cool.TimeToAct())
			for (int i = 0; i<4; i++)
				SpawnDrone(RandomBorderPosition((1<<1) ^ (1<<3)));
	}

	if (gametime > 65.0 && gametime < 80.0) {
		static Cooldown cool(0.4, 0.0);
		cool.PassTime(deltatime);
		if (cool.TimeToAct()) {
			SpawnShooter(RandomBorderPosition(1<<2));
        }

		static Cooldown pcon(1.0, 0.0);
		pcon.PassTime(deltatime);
		if (pcon.TimeToAct()) {
			ControlPopulation(MAX_MONSTER_COUNT - 5);
		}

	}

	if (gametime > 85.0 && gametime < 90.0) {	//one-time spawn
		static int k = 5;
		static Cooldown cool(1000.0, 0.0);
		cool.PassTime(deltatime);
		if (cool.TimeToAct()) {
			ControlPopulation(MAX_MONSTER_COUNT - k);
			for (int i = 0; i<k; i++) SpawnDodger(RandomBorderPosition());
		}
	}

   	if (gametime > 90.0 && gametime < 110.0) {
		static int k = 3;
		static Cooldown cool(1.0, 0.0);
		cool.PassTime(deltatime);
		if (cool.TimeToAct()) {
			ControlPopulation(MAX_MONSTER_COUNT - k);
			for (int i = 0; i<k; i++)
				SpawnDrone(RandomBorderPosition());
		}
	}
   	if (gametime > 110.0 && gametime < 120.0) {
		static int k = 20;
		static Cooldown cool(1.0, 0.0);
		cool.PassTime(deltatime);
		if (cool.TimeToAct()) {
			ControlPopulation(MAX_MONSTER_COUNT - k);
			for (int i = 0; i<k; i++) {
				int m = (1<<0) | (1<<1) | (1<<2);
				if (trandom() < 0.03) SpawnShooter(RandomBorderPosition(m));
				SpawnDrone(RandomBorderPosition(m));
			}
		}
	}
   	if (gametime > 125.0 && gametime < 160.0) {
		static Cooldown cwall(10000.0, 0.0);	//one-time spawn
		cwall.PassTime(deltatime);
		if (cwall.TimeToAct()) SpawnWall(0);

		static int k = 30;
		static Cooldown cool(5.0, 0.0);
		cool.PassTime(deltatime);
		if (cool.TimeToAct()) {
			for (int i = 0; i<k; i++) {
				if (trandom() < 0.01) SpawnShooter(RandomBorderPosition());
				SpawnDrone(RandomBorderPosition());
			}
			k += 5;
		}
	}
   	if (gametime > 160.0 && gametime < 190.0) {
		static Cooldown cwall(10000.0, 0.0);	//one-time spawn
		cwall.PassTime(deltatime);
		if (cwall.TimeToAct()) {
			int k = 3;
			ControlPopulation(MAX_MONSTER_COUNT - k);
			for (int i = 0; i<k; i++) SpawnDodger(RandomBorderPosition());
			SpawnWall(1);
			SpawnWall(2);
		}

		static int k = 5;
		static Cooldown cool(1.0, 0.0);
		cool.PassTime(deltatime);
		if (cool.TimeToAct()) {
			for (int i = 0; i<k; i++) {
				if (trandom() < 0.01) SpawnShooter(RandomBorderPosition());
				SpawnDrone(RandomBorderPosition());
			}
		}
	}
   	if (gametime > 190.0 && gametime < 191.0) {
		static Cooldown cwall(10000.0, 0.0);	//one-time spawn
		cwall.PassTime(deltatime);
		if (cwall.TimeToAct()) {
			SpawnWall(0);
			SpawnWall(1);
			SpawnWall(2);
			SpawnWall(3);
		}
	}
	//Here the infinite wave begins
   	if (gametime > 210.0) {
		static Cooldown cwall(20.0, 0.0);
		static int q = 2;
		cwall.PassTime(deltatime);
		if (cwall.TimeToAct()) {
			for (int i = 0; i<q; i++) SpawnWall(rand() % 4);
			q *= 2;
		}

		double t = (gametime - 210.0) / (300.0 - 210.0);
		double pbad = exp(log(100.0) * (t - 1.0));

		static Cooldown cool(1.0, 0.0);
		static int k = 10;
		cool.PassTime(deltatime);
		if (cool.TimeToAct()) {
			ControlPopulation(MAX_MONSTER_COUNT - k);
			for (int i = 0; i<k; i++) {
				if (trandom() < pbad) {
					double rnd = trandom();
					rnd -= 0.65;  if (rnd < 0) SpawnShooter(RandomBorderPosition());
					rnd -= 0.30;  if (rnd < 0) SpawnSpeedy(RandomBorderPosition());
					rnd -= 0.05;  if (rnd < 0) SpawnDodger(RandomBorderPosition());
				}
				else SpawnDrone(RandomBorderPosition());
			}
		}
   	}


/*	//5 guys are enough for thrilling dodging experience
	static int k = 5;
	static Cooldown cool(10.0, 0.0);
	cool.PassTime(deltatime);
	if (cool.TimeToAct()) {
		ControlPopulation(MAX_MONSTER_COUNT - k);
		for (int i = 0; i<k; i++) {
			SpawnDodger(RandomBorderPosition());
		}
	}*/

/*	//25 seconds is enough to kill dull
	//shoot on moving is much better with it
	static Cooldown cool(0.5, 0.0);
	cool.PassTime(deltatime);
	if (cool.TimeToAct()) {
		SpawnSpeedy(RandomDirection());
	}*/

/*	//50 secs: separation between dull and cool moving tactic
	static int k = 50;
	static Cooldown cool(10.0, 0.0);
	cool.PassTime(deltatime);
	if (cool.TimeToAct()) {
		ControlPopulation(MAX_MONSTER_COUNT - k);
		for (int i = 0; i<k; i++) {
			SpawnDrone(RandomBorderPosition());
		}
		k += 30;
	}*/

/*	static Cooldown cwall(30.0);
	cwall.PassTime(deltatime);
	if (cwall.TimeToAct()) {
		SpawnWall(rand()%4);
	}*/
}
