#include "Entities.h"
#include "GameWorld.h"

MovingCircle Entity::GetGeometry() {
	return MovingCircle(position, newposition, radius);
}

void Entity::Move(double deltatime) {
	newposition = position + velocity * deltatime;
}

void Missile::DealDamage(Entity *ent) {
	double dam = std::min(damage, hitpoints + EPS);
	if (dam < 0.0) dam = 0.0;
	ent->hitpoints -= dam;
	hitpoints -= dam;

	//If is it player's deadly hit -> set HP := -inf
	if (ent->hitpoints <= 0.0 && classname == 0) ent->hitpoints = -INF;

	ent->velocity *= MISSILE_IMPULSE_STOP;
	ent->velocity += ForceVectorNorm(velocity, MISSILE_IMPULSE_VELOCITY);
}

void Guy::SetMeleeWeapon(MeleeWeapon *_teeth) {
	teeth = std::auto_ptr<MeleeWeapon>(_teeth);
}

void Guy::SetRangedWeapon(RangedWeapon *_gun) {
	gun = std::auto_ptr<RangedWeapon>(_gun);
}

void Guy::PassTime(double deltatime) {
	Entity::PassTime(deltatime);

	if (teeth.get()) teeth->cooldown.PassTime(deltatime);
	if (gun.get()) gun->cooldown.PassTime(deltatime);
}

void Guy::Move(double deltatime) {
	desiredvelocity = LimitVectorNorm(desiredvelocity, maxvelocity);
	double2 deltavel = desiredvelocity - velocity;
	deltavel = LimitVectorNorm(deltavel, maxacceleration * deltatime);
	velocity += deltavel;
	Entity::Move(deltatime);
}

void Guy::Shoot(double2 direction) {
	if (!gun.get()) return;
	Missile *miss = gun->Shoot(position, direction);
	if (!miss) return;
	miss->classname = classname;
	miss->nowhit.push_back(this);
	owner->BindEntity(owner->AllocateMissile(), miss);
}

void Guy::SetMoveDirection(double2 direction) {
	desiredvelocity = ForceVectorNorm(direction, maxvelocity);
}

void Monster::SetAI(AIBase *_ai) {
	ai = std::auto_ptr<AIBase>(_ai);
}

void Monster::PassTime(double deltatime) {
	Guy::PassTime(deltatime);

	if (ai.get()) ai->PassTime(deltatime);
}
