#ifndef GAME_ENTITIES_687318
#define GAME_ENTITIES_687318
//everything that is considered to be game objects
//don't know what to do with events - I don't need them...

//#define U2C "Unknown to contestants."

#include "Global.h"
#include "GeomObjects.h"
#include "AI.h"
struct GameWorld;

struct Entity {
	//index in the entities array
	int ID;						//U2C
	//class of object (different for different types of monsters/missiles)
	int classname;
	//pointer to GameWorld (required in AI sometimes)
	GameWorld *owner;			//U2C

	//current position
	double2 position;
	//position in the next frame
	double2 newposition;		//U2C

	//current velocity vector
	double2 velocity;

	//circle of radius
	double radius;
	//number of hit points
	double hitpoints;
	//maximal number of hit points
	const double maxhitpoints;

	inline Entity(int _classname, double2 _position, double _radius, double _hitpoints)
		 : ID(-1), classname(_classname), owner(0)
		 , position(_position), newposition(_position)
		 , velocity()
		 , radius(_radius), hitpoints(_hitpoints), maxhitpoints(_hitpoints) {}
	inline virtual ~Entity() {}

	MovingCircle GetGeometry();

	inline virtual void PassTime(double deltatime) {}

	virtual void Move(double deltatime);

};

struct Missile : public Entity {
	//damage per hit
	double damage;
	//the entities that are colliding now
	std::vector<Entity *> nowhit;	//U2C

	inline Missile(double2 _position, int _classname, double _radius, double _hitpoints, double _damage)
		 : Entity(_classname, _position, _radius, _hitpoints)
		 , damage(_damage), nowhit() {}

	void DealDamage(Entity *ent);
};

struct Guy : public Entity {
	//the goal velocity
	double2 desiredvelocity;		//U2C
	//maximum possible velocity
	double maxvelocity;
	//maximum possible acceleration
	double maxacceleration;

	//melee weapon (if any)
	std::auto_ptr<MeleeWeapon> teeth;
	//ranged weapon (if any)
	std::auto_ptr<RangedWeapon> gun;


	inline Guy(double2 _position, int _classname, double _radius, double _hitpoints, double _maxvelocity, double _maxacceleration)
		 : Entity(_classname, _position, _radius, _hitpoints)
		 , desiredvelocity(), maxvelocity(_maxvelocity), maxacceleration(_maxacceleration)
		 , teeth(0), gun(0) {}
	void SetMeleeWeapon(MeleeWeapon *_teeth);
	void SetRangedWeapon(RangedWeapon *_gun);
	inline virtual ~Guy() {}


	virtual void PassTime(double deltatime);

	virtual void Move(double deltatime);

	void Shoot(double2 direction);

	void SetMoveDirection(double2 direction);
};

struct Player : public Guy {

	inline Player(double2 _position, int _classname, double _radius, double _hitpoints, double _maxvelocity, double _maxacceleration)
		 : Guy(_position, _classname, _radius, _hitpoints, _maxvelocity, _maxacceleration) {}
};

struct Monster : public Guy {
	//AI object
	std::auto_ptr<AIBase> ai;		//U2C

	inline Monster(double2 _position, int _classname, double _radius, double _hitpoints, double _maxvelocity, double _maxacceleration)
		 : Guy(_position, _classname, _radius, _hitpoints, _maxvelocity, _maxacceleration) {}
	void SetAI(AIBase *_ai);

	void PassTime(double deltatime);

};

#endif
