#ifndef GAME_WORLD_873218
#define GAME_WORLD_873218
//all game state must be inside GameWorld
//cpp contains the main function implementation

#include "Global.h"
#include "Entities.h"
#include "Grid.h"
#include "Director.h"

template<class E, int S> struct EntityArray {
	E arr[S];

	inline int size() const { return S; }
	inline E &operator[] (int i) { return arr[i]; }
	inline const E &operator[] (int i) const { return arr[i]; }
};

//typeinfo for entities
bool IsPlayer(int &ID);
bool IsMonster(int &ID);
bool IsMissile(int &ID);
bool IsPlayer(const Entity *ent);
bool IsMonster(const Entity *ent);
bool IsMissile(const Entity *ent);

struct GameWorld {
	//array of entities (index = ID)
	static const int MAX_ENTITY_COUNT = MAX_PLAYER_COUNT + MAX_MONSTER_COUNT + MAX_MISSILE_COUNT;
	EntityArray<Entity *, MAX_ENTITY_COUNT> entities;

	//some ptrs for convenience
	Player **players;
	Monster **monsters;
	Missile **missiles;

	//collision detection optimization
	Grid grid;

	//player shoot control
	bool playerwantstoshoot;
	double2 playershotdirection;

	//director controls player/monster creation on the field
	std::auto_ptr<Director> director;
	//number of frags;
	int frags;

	//construct the world 
	//director is created automatically
	GameWorld();
	~GameWorld();

	//the main function
	//	deltatime = time to simulate
	//	return = error code (death, playing, fail, ...)
    int Iterate(double deltatime);
    enum IterationResult {
		SUCCESS = 0,
		DEATH,
    };

    //use these methods to set player's behaviour
    void SetPlayerMove(double2 direction);
    void SetPlayerShoot(double2 direction);

    //get the contestant's score
    int GetScore();

	//allocate entity IDs
	int AllocateMonster();
	int AllocateMissile();

	//bind entity to ID
	Entity *BindEntity(int ID, Entity *ent);
	//unbind and destroy entity
	void DestroyEntity(Entity *ent);

private:
	typedef std::vector<Entity *> EntitySet;
    void FillGrid();
	EntitySet FindCollision(const MovingCircle &go);
};

#endif
