#include "Serialization.h"

static const int ENTITY_DELTA_CLASSNAME = 100;
static const int DEAD_CLASSNAME = -1;

#define SERIALIZE_IFDEAD(res, ent) {\
	if (!ent) {\
		res.ClassName = DEAD_CLASSNAME;\
		return;\
	}\
}

#define SERIALIZE_ENTITY(res, ent) {\
	res.PositionX = ent->position.x;\
	res.PositionY = ent->position.y;\
	res.VelocityX = ent->velocity.x;\
	res.VelocityY = ent->velocity.y;\
	res.Radius = ent->radius;\
	res.HitPoints = ent->hitpoints;\
	res.MaxHitPoints = ent->maxhitpoints;\
}

#define SERIALIZE_MISSILE(res, entity) {\
	const Missile *ent = (const Missile *)(entity);\
	res.Damage = ent->damage;\
}

#define SERIALIZE_GUY(res, entity) {\
	const Guy *ent = (const Guy *)(entity);\
	res.MaxVelocity = ent->maxvelocity;\
	res.MaxAcceleration = ent->maxacceleration;\
}

#define SERIALIZE_RANGEDWEAPON(res, entity) {\
	const Guy *ent = (const Guy *)(entity);\
    RangedWeapon *wrng = ent->gun.get();\
    if (wrng) {\
		res.GunCooldownRemaining = wrng->cooldown.remaining;\
		res.GunCooldownPeriod = wrng->cooldown.period;\
		res.BulletDamage = wrng->bulletdamage;\
		res.BulletHitPoints = wrng->bullethitpoints;\
		res.BulletVelocity = wrng->bulletvelocity;\
		res.BulletRadius = wrng->bulletradius;\
	}\
}

#define SERIALIZE_MELEEWEAPON(res, entity) {\
	const Guy *ent = (const Guy *)(entity);\
	MeleeWeapon *wmel = ent->teeth.get();\
    if (wmel) {\
		res.TeethCooldownRemaining = wmel->cooldown.remaining;\
		res.TeethCooldownPeriod = wmel->cooldown.period;\
		res.BiteDamage = wmel->bitedamage;\
	}\
}

void SerializePlayer(CPlayer &res, const Entity *entity) {
	SERIALIZE_IFDEAD(res, entity);
	res.ClassName = entity->classname;
	SERIALIZE_ENTITY(res, entity);
	SERIALIZE_GUY(res, entity);
	SERIALIZE_RANGEDWEAPON(res, entity);
}

void SerializeMonster(CMonster &res, const Entity *entity) {
	SERIALIZE_IFDEAD(res, entity);
	res.ClassName = entity->classname + ENTITY_DELTA_CLASSNAME;
	SERIALIZE_ENTITY(res, entity);
	SERIALIZE_GUY(res, entity);
	SERIALIZE_RANGEDWEAPON(res, entity);
	SERIALIZE_MELEEWEAPON(res, entity);
}

void SerializeMissile(CMissile &res, const Entity *entity) {
	SERIALIZE_IFDEAD(res, entity);
	res.ClassName = entity->classname + 2*ENTITY_DELTA_CLASSNAME;
	SERIALIZE_ENTITY(res, entity);
	SERIALIZE_MISSILE(res, entity);
}

void SerializeWorld(World &world, const GameWorld *gameworld) {
	memset(world.monsters, 0, sizeof(world.monsters));
	memset(world.missiles, 0, sizeof(world.missiles));
	for (int i = 0; i<gameworld->entities.size(); i++) {
		const Entity *ent = gameworld->entities[i];
		int ID;

		ID = i;  if (IsPlayer(ID)) SerializePlayer(world.player, ent);
		ID = i;  if (IsMonster(ID)) SerializeMonster(world.monsters[ID], ent);
		ID = i;  if (IsMissile(ID)) SerializeMissile(world.missiles[ID], ent);
	}
}
