#include "../cpp/world.h"
#include <cstdio>
#include <cstring>
#include <cmath>

static const double EPS = 1e-9;

struct Point {
	double x, y;
	
	inline Point() : x(0), y(0) {}
	inline Point(double _x, double _y) : x(_x), y(_y) {}

	inline       double &operator[] (int i)       { return (i==0 ? x : y); }
	inline const double &operator[] (int i) const { return (i==0 ? x : y); }

	inline Point operator+ (const Point &b) const { return Point(x+b.x, y+b.y); }
	inline Point operator- (const Point &b) const { return Point(x-b.x, y-b.y); }
	inline Point operator* (double b) const { return Point(x*b, y*b); }
	inline Point operator/ (double b) const { return *this * (1.0/b); }

	inline void operator+= (const Point &b) { *this = *this + b; }
	inline void operator-= (const Point &b) { *this = *this - b; }
	inline void operator*= (double b) { *this = *this * b; }
	inline void operator/= (double b) { *this = *this / b; }

	inline double operator^ (const Point &b) const { return x*b.x + y*b.y; }
	inline double operator% (const Point &b) const { return x*b.y - y*b.x; }

	inline double operator!() const { return sqrt(*this ^ *this); }
};

Point Normal(const Point &x) {
	Point res = x;
	double norm = !res;
	if (norm > EPS) res /= norm;
	return res;
}

#define POINT(name) Point(name##X, name##Y)
#define COORDS(pnt) (pnt).x, (pnt.y)
static const int MONSTERS = World::MAXMONSTERS;
static const int MISSILES = World::MAXMISSILES;

extern World *world;



static const double MOVE_RADIUS = 0.3;

Point TargetMonsterDull(const CMonster &monster) {
	return POINT(monster.Position) - POINT(world->player.Position);
}

void BotThink() {
	//Targeting
	double mind = 1e+50;
	Point best;
	for (int i = 0; i<MONSTERS; i++) if (world->monsters[i].ClassName >= 0) {
		Point target = TargetMonsterDull(world->monsters[i]);
		if (mind > !target) {
			best = target;
			mind = !best;
		}
	}
	world->shoot(COORDS(best));

	//Moving
	Point radiusvec = POINT(world->player.Position) - Point(0.5, 0.5);
	double dist = !radiusvec;
	if (dist < EPS) radiusvec = Point(1, 0);
	else radiusvec = Normal(radiusvec);
	Point tangentvec(-radiusvec.y, radiusvec.x);

	Point force = tangentvec + radiusvec * (MOVE_RADIUS - dist) / MOVE_RADIUS;
	world->run(COORDS(force));
}

#ifndef BOTWITHGLUT
World *world;
int main(int argc, char** argv) {
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 666;
	}
	world = new World(argv[1], atoi(argv[2]));
	world->makeMove();
	while (1) {
		BotThink();
		world->makeMove();
	}
	printf("Game over!\n");
	delete world;
	return 0;
}
#endif